@@ -6,14 +6,17 @@ import time, json
 
 # Pip
 from selenium_firefox.firefox import Firefox, By, Keys
-from ktimeout import timeout
+from selenium.webdriver.common.action_chains import ActionChains
+import stopit
 from kcu import strings
 
 from bs4 import BeautifulSoup as bs
 
 # Local
-from .visibility import Visibility
-from .upload_status import UploadStatus
+from .enums.visibility import Visibility
+from .enums.upload_status import UploadStatus
+from .enums.analytics_period import AnalyticsPeriod
+from .enums.analytics_tab import AnalyticsTab
 
 # ---------------------------------------------------------------------------------------------------------------------------------------- #
 
@@ -26,12 +29,15 @@ YT_STUDIO_URL   = 'https://studio.youtube.com'
 YT_UPLOAD_URL   = 'https://www.youtube.com/upload'
 YT_LOGIN_URL    = 'https://accounts.google.com/signin/v2/identifier?service=youtube'
 YT_STUDIO_VIDEO_URL = 'https://studio.youtube.com/video/{}/edit/basic'
+YT_WATCH_VIDEO_URL = 'https://www.youtube.com/watch?v='
 
 MAX_TITLE_CHAR_LEN          = 100
 MAX_DESCRIPTION_CHAR_LEN    = 5000
-MAX_TAGS_CHAR_LEN           = 500
+MAX_TAGS_CHAR_LEN           = 400
+MAX_TAG_CHAR_LEN            = 30
 
 LOGIN_INFO_COOKIE_NAME = 'LOGIN_INFO'
+TIME_OUT_ERROR = 'Operation has timed out.'
 
 # ---------------------------------------------------------------------------------------------------------------------------------------- #
 
@@ -45,8 +51,8 @@ class Youtube:
 
     def __init__(
         self,
-        cookies_folder_path: str,
-        extensions_folder_path: str,
+        cookies_folder_path: Optional[str] = None,
+        extensions_folder_path: Optional[str] = None,
         email: Optional[str] = None,
         password: Optional[str] = None,
         host: Optional[str] = None,
@@ -55,24 +61,34 @@ class Youtube:
         full_screen: bool = False,
         disable_images: bool = False,
         user_agent: Optional[str] = None,
-        login_prompt_callback: Optional[Callable[[str], None]] = None
+        login_prompt_callback: Optional[Callable[[str], None]] = None,
+        headless: bool = False
     ):
-        self.browser = Firefox(cookies_folder_path, extensions_folder_path, host=host, port=port, screen_size=screen_size, user_agent=user_agent, full_screen=full_screen, disable_images=disable_images)
+        self.browser = Firefox(
+            cookies_folder_path, extensions_folder_path, host=host, port=port, 
+            screen_size=screen_size, user_agent=user_agent, 
+            full_screen=full_screen, disable_images=disable_images,
+            headless=headless
+        )
         self.channel_id = None
+        self.browser.get(YT_URL)
+        self.__dismiss_alerts()
 
         try:
             self.__internal_channel_id = cookies_folder_path.strip('/').split('/')[-1]
         except:
-            self.__internal_channel_id = cookies_folder_path
+            self.__internal_channel_id = cookies_folder_path or self.browser.cookies_folder_path
 
         try:
-            if self.browser.login_via_cookies(YT_URL, LOGIN_INFO_COOKIE_NAME) or self.login(email=email, password=password, login_prompt_callback=login_prompt_callback):
+            if self.browser.login_via_cookies(YT_URL, LOGIN_INFO_COOKIE_NAME):
                 time.sleep(0.5)
                 self.browser.get(YT_URL)
                 time.sleep(0.5)
                 self.browser.save_cookies()
                 time.sleep(0.5)
                 self.channel_id = self.get_current_channel_id()
+            elif email and password:
+                self.login(email=email, password=password, login_prompt_callback=login_prompt_callback)
         except Exception as e:
             print(e)
             self.quit()
@@ -112,14 +128,80 @@ class Youtube:
             time.sleep(0.5)
             self.browser.get(YT_URL)
             time.sleep(0.5)
-
             self.browser.save_cookies()
+            time.sleep(0.5)
+            self.channel_id = self.get_current_channel_id()
 
         time.sleep(0.5)
         self.browser.get(org_url)
 
         return logged_in
 
+    def watch_video(
+        self,
+        video_id: str,
+        percent_to_watch: float = -1, # 0-100 # -1 means all
+        like: bool = False
+    ) -> Tuple[bool, bool]: # watched, liked
+        watched = False
+        liked = False
+
+        try:
+            self.browser.get(YT_WATCH_VIDEO_URL+video_id)
+            length_s = float(strings.between(self.browser.driver.page_source, 'detailpage\\\\u0026len=', '\\\\'))
+            play_button = self.browser.find_by('button', class_='ytp-large-play-button ytp-button', timeout=0.5)
+
+            if play_button and play_button.is_displayed():
+                play_button.click()
+                time.sleep(1)
+
+            while True:
+                ad = self.browser.find_by('div', class_='video-ads ytp-ad-module', timeout=0.5)
+
+                if not ad or not ad.is_displayed():
+                    break
+
+                time.sleep(0.1)
+
+            watched = True
+            seconds_to_watch = percent_to_watch / 100 * length_s if percent_to_watch >= 0 else length_s
+
+            if seconds_to_watch > 0:
+                print('Goinng to watch', seconds_to_watch)
+                time.sleep(seconds_to_watch)
+
+            return watched, self.like(video_id) if like and self.is_logged_in else False
+        except Exception as e:
+            print(e)
+
+            return watched, liked
+    
+    def like(self, video_id: str) -> bool:
+        self.browser.get(YT_WATCH_VIDEO_URL+video_id)
+
+        try:
+            buttons_container = self.browser.find_by('div', id_='top-level-buttons', class_='style-scope ytd-menu-renderer', timeout=1.5)
+
+            if buttons_container:
+                button_container = self.browser.find_by('ytd-toggle-button-renderer', class_='style-scope ytd-menu-renderer force-icon-button style-text', timeout=0.5, in_element=buttons_container)
+
+                if button_container:
+                    button = self.browser.find_by('button', id_='button', timeout=0.5, in_element=button_container)
+
+                    if button:
+                        attr = button.get_attribute('aria-pressed')
+
+                        if attr and attr == 'false':
+                            button.click()
+
+                        return True
+
+            return False
+        except Exception as e:
+            print(e)
+
+            return False
+
     def upload(
         self,
         video_path: str,
@@ -133,32 +215,43 @@ class Youtube:
         extra_sleep_after_upload: Optional[int] = None,
         extra_sleep_before_publish: Optional[int] = None
     ) -> (bool, Optional[str]):
+        res = self.__upload(video_path, title, description, tags, made_for_kids=made_for_kids, visibility=visibility, thumbnail_image_path=thumbnail_image_path, extra_sleep_after_upload=extra_sleep_after_upload, extra_sleep_before_publish=extra_sleep_before_publish, timeout=_timeout)
 
-        if _timeout is not None:
-            try:
-                with timeout.timeout(_timeout):
-                    return self.__upload(video_path, title, description, tags, made_for_kids=made_for_kids, visibility=visibility, thumbnail_image_path=thumbnail_image_path, extra_sleep_after_upload=extra_sleep_after_upload, extra_sleep_before_publish=extra_sleep_before_publish)
-            except Exception as e:
-                print('Upload', e)
-                # self.browser.get(YT_URL)
+        if res == TIME_OUT_ERROR:
+            print('Upload:', TIME_OUT_ERROR)
 
-                return False, None
-        else:
-            return self.__upload(video_path, title, description, tags, made_for_kids=made_for_kids, visibility=visibility, thumbnail_image_path=thumbnail_image_path, extra_sleep_after_upload=extra_sleep_after_upload, extra_sleep_before_publish=extra_sleep_before_publish)
+            return False, None
 
-    def get_current_channel_id(self) -> Optional[str]:
-        self.browser.get(YT_URL)
+        return res
+
+    def get_current_channel_id(self, _click_avatar: bool = False, _get_home_url: bool = False) -> Optional[str]:
+        if _get_home_url:
+            self.browser.get(YT_URL)
 
         try:
-            return json.loads(
-                strings.between(
-                    self.browser.driver.page_source, 'var ytInitialGuideData = ', '};'
-                ) + '}'
-            )['responseContext']['serviceTrackingParams'][2]['params'][0]['value']
-        except Exception as e:
-            print('get_current_channel_id', e)
+            if _click_avatar:
+                avatar_button = self.browser.find_by('button', id_='avatar-btn', timeout=0.5)
 
-            return None
+                if avatar_button:
+                    avatar_button.click()
+
+            href_containers = self.browser.find_all_by('a', class_='yt-simple-endpoint style-scope ytd-compact-link-renderer', timeout=0.5)
+
+            if href_containers:
+                for href_container in href_containers:
+                    href = href_container.get_attribute('href')
+
+                    if href and 'channel/' in href:
+                        return strings.between(href, 'channel/', '?')
+        except Exception as e:
+            print(e)
+
+        if not _click_avatar:
+            return self.get_current_channel_id(_click_avatar=True, _get_home_url=_get_home_url)
+        elif not _get_home_url:
+            return self.get_current_channel_id(_click_avatar=False, _get_home_url=True)
+
+        return None
 
     def load_video(self, video_id: str):
         self.browser.get(self.__video_url(video_id))
@@ -170,17 +263,14 @@ class Youtube:
         pinned: bool = False,
         _timeout: Optional[int] = 15
     ) -> (bool, bool):
-        if _timeout is not None:
-            try:
-                with timeout.timeout(_timeout):
-                    return self.__comment_on_video(video_id, comment, pinned=pinned)
-            except Exception as e:
-                print('Comment', e)
-                # self.browser.get(YT_URL)
+        res = self.__comment_on_video(video_id, comment, pinned=pinned, timeout=_timeout)
 
-                return False, False
-        else:
-            return self.__comment_on_video(video_id, comment, pinned=pinned)
+        if res == TIME_OUT_ERROR:
+            print('Comment:', TIME_OUT_ERROR)
+
+            return False, False
+
+        return res
 
     def get_channel_video_ids(
         self,
@@ -244,11 +334,20 @@ class Youtube:
 
         return video_ids
 
-    def check_analytics(self) -> bool:
-        self.browser.get(YT_STUDIO_URL)
+    def check_analytics(
+        self,
+        tab: AnalyticsTab = AnalyticsTab.OVERVIEW,
+        period: AnalyticsPeriod = AnalyticsPeriod.LAST_28_DAYS
+    ) -> bool:
+        if not self.channel_id:
+            print('No channel ID found')
+
+            return False
+
+        url = YT_STUDIO_URL.rstrip('/') + '/channel/' + self.channel_id + '/analytics/tab-' + tab.value + '/period-' + period.value
 
         try:
-            self.browser.get(self.browser.find(By.XPATH, "//a[@id='menu-item-3']").get_attribute('href'))
+            self.browser.get(url)
 
             return True
         except Exception as e:
@@ -256,6 +355,34 @@ class Youtube:
 
             return False
 
+    def get_violations(self) -> Tuple[bool, int]: # has_warning, strikes
+        self.browser.get(YT_STUDIO_URL)
+
+        try:
+            violations_container = self.browser.find_by('div', class_='style-scope ytcd-strikes-item')
+
+            if not violations_container:
+                return False, 0
+
+            violations_label = self.browser.find_by('div', class_='label style-scope ytcp-badge', in_element=violations_container)
+
+            if not violations_label:
+                return False, 0
+
+            violation_text = violations_label.text.strip().lower()
+            violation_text_number = 0
+
+            try:
+                violation_text_number = int(violation_text)
+            except:
+                pass
+
+            return True, violation_text_number
+        except Exception as e:
+            print(e)
+
+            return False, 0
+
     def add_endscreen(self, video_id: str, max_wait_seconds_for_processing: float = 0) -> bool:
         self.browser.get(YT_STUDIO_VIDEO_URL.format(video_id))
 
@@ -298,6 +425,7 @@ class Youtube:
 
     # ------------------------------------------------------- Private methods -------------------------------------------------------- #
 
+    @stopit.threading_timeoutable(default=TIME_OUT_ERROR, timeout_param='timeout')
     def __upload(
         self,
         video_path: str,
@@ -308,7 +436,8 @@ class Youtube:
         visibility: Visibility = Visibility.PUBLIC,
         thumbnail_image_path: Optional[str] = None,
         extra_sleep_after_upload: Optional[int] = None,
-        extra_sleep_before_publish: Optional[int] = None
+        extra_sleep_before_publish: Optional[int] = None,
+        timeout: Optional[int] = None
     ) -> (bool, Optional[str]):
         self.browser.get(YT_URL)
         time.sleep(1.5)
@@ -352,7 +481,7 @@ class Youtube:
 
             tags_container = self.browser.find(By.XPATH, "/html/body/ytcp-uploads-dialog/paper-dialog/div/ytcp-animatable[1]/ytcp-uploads-details/div/ytcp-uploads-advanced/ytcp-form-input-container/div[1]/div[2]/ytcp-free-text-chip-bar/ytcp-chip-bar/div")
             tags_field = self.browser.find(By.ID, 'text-input', tags_container)
-            tags_field.send_keys(','.join(tags) + ',')
+            tags_field.send_keys(','.join([t for t in tags if len(t) <= MAX_TAG_CHAR_LEN])[:MAX_TAGS_CHAR_LEN-1] + ',')
             print("Upload: added tags")
 
             kids_selection_name = 'MADE_FOR_KIDS' if made_for_kids else 'NOT_MADE_FOR_KIDS'
@@ -433,7 +562,8 @@ class Youtube:
             return False, None
 
     # returns (commented_successfully, pinned_comment_successfully)
-    def __comment_on_video(self, video_id: str, comment: str, pinned: bool = False) -> (bool, bool):
+    @stopit.threading_timeoutable(default=TIME_OUT_ERROR, timeout_param='timeout')
+    def __comment_on_video(self, video_id: str, comment: str, pinned: bool = False, timeout: Optional[int] = None) -> (bool, bool):
         self.load_video(video_id)
         time.sleep(1)
         self.browser.scroll(150)
@@ -549,7 +679,6 @@ class Youtube:
                 self.browser.find_by('div', {'jscontroller': 'VXdfxd', 'role': 'button'}).click()
                 time.sleep(1)
 
-                from selenium.webdriver.common.action_chains import ActionChains
                 action = ActionChains(self.browser.driver)
 
                 pass_container = self.browser.find_by('div', { 'id': 'password', 'jscontroller': 'pxq3x' })
@@ -585,6 +714,28 @@ class Youtube:
 
         return self.is_logged_in
     
+    def __dismiss_alerts(self):
+        dismiss_button_container = self.browser.find_by('div', id_='dismiss-button', timeout=1.5)
+        
+        if dismiss_button_container:
+            dismiss_button = self.browser.find_by('paper-button', id_='button', timeout=0.5, in_element=dismiss_button_container)
+
+            if dismiss_button:
+                dismiss_button.click()
+            
+            iframe = self.browser.find_by('iframe', class_='style-scope ytd-consent-bump-lightbox', timeout=2.5)
+
+            if iframe:
+                self.browser.driver.switch_to.frame(iframe)
+
+            agree_button = self.browser.find_by('div', id_='introAgreeButton', timeout=2.5)
+
+            if agree_button:
+                agree_button.click()
+            
+            if iframe:
+                self.browser.driver.switch_to.default_content()
+    
     def __video_url(self, video_id: str) -> str:
         return YT_URL + '/watch?v=' + video_id
     
@@ -592,4 +743,4 @@ class Youtube:
         return YT_URL + '/channel/' + channel_id + '/videos?view=0&sort=da&flow=grid'
 
 
-# ---------------------------------------------------------------------------------------------------------------------------------------- #
\ No newline at end of file
+# ---------------------------------------------------------------------------------------------------------------------------------------- #s
\ No newline at end of file
